import ReconnectingWebSocket from 'reconnecting-websocket'
import { MessageTypes, numberToUint16BE, packers, unpackers, unpackIncomingPacket } from 'bennc-js'
import { useMercuryStore } from '../stores/mercuryStore'
import { useUserStore } from '../stores/userStore'
import { useMessageStore } from '../stores/messageStore'
import { useChannelStore } from '../stores/channelStore'
import { decrypt } from 'romulus-js'
import { provide } from 'vue'

const CLIENT_ID = 'Mercury'
const SUBSCRIBED_MESSAGE_TYPES = [MessageTypes.Basic, MessageTypes.UserDataRequest, MessageTypes.UserDataResponse]

const encoder = new TextEncoder()
const decoder = new TextDecoder()

export default {
  install: (app, options) => {
    // Create websocket.
    const ws = new ReconnectingWebSocket('wss://chat.3t.network/bennc')
    ws.binaryType = 'arraybuffer'
    app.provide('websocket', ws)

    // Create store accesses.
    const mercuryStore = useMercuryStore()
    const channelStore = useChannelStore()
    const messageStore = useMessageStore()
    const userStore = useUserStore()


    function handleBasicMessage(incomingPacket) { 
      // Try decryption with each encryption key.
      for (let channel of channelStore.channels) {
        // Attempt decryption of incoming message.
        const result = decrypt(incomingPacket.data, numberToUint16BE(MessageTypes.Basic), channel.key.raw)

        // If encryption key is found...
        if (result.success) {
          const unpackedData = unpackers[MessageTypes.Basic](result.plaintext)

          // Add channel to user's channel list.
          userStore.addChannel(incomingPacket.senderId, channel.id)

          // Add message to message store.
          messageStore.addMessage(channel.id, incomingPacket.senderId, decoder.decode(unpackedData))

          // Don't process other keys.
          break
        }
      }
    }

    function handleUserDataRequest(incomingPacket) {
      // Try decryption with each encryption key.
      for (let channel of channelStore.channels) {
        // Attempt decryption of incoming message.
        const result = decrypt(incomingPacket.data, numberToUint16BE(MessageTypes.UserDataRequest), channel.key.raw)

        // If encryption key is found...
        if (result.success) {
          // Unpack user data request.
          const userDataRequest = unpackers[MessageTypes.UserDataRequest](result.plaintext)

          // Update user in user store.
          userStore.setName(incomingPacket.senderId, userDataRequest.username)
          userStore.setColor(incomingPacket.senderId, userDataRequest.colour)
          userStore.addChannel(incomingPacket.senderId, channel.id)

          // Respond to user data request with a user data response.
          const self = userStore.getSelf()
          const userDataResponse = packers[MessageTypes.UserDataResponse]({ username: self.name, colour: self.color, clientId: CLIENT_ID }, channel.key.raw)
          ws.send(userDataResponse)

          // Don't process other keys.
          break
        }
      }
    }

    function handleUserDataResponse(incomingPacket) {
      // Try decryption with each encryption key.
      for (let channel of channelStore.channels) {
        // Attempt decryption of incoming message.
        const result = decrypt(incomingPacket.data, numberToUint16BE(MessageTypes.UserDataResponse), channel.key.raw)
        
        // If encryption key is found...
        if (result.success) {
          // Unpack user data response.
          const userDataResponse = unpackers[MessageTypes.UserDataResponse](result.plaintext)

          // Update user in user store.
          userStore.setName(incomingPacket.senderId, userDataResponse.username)
          userStore.setColor(incomingPacket.senderId, userDataResponse.colour)
          userStore.addChannel(incomingPacket.senderId, channel.id)

          // Don't process other keys.
          break
        }
      }
    }

    function requestUserData() {
      const self = userStore.getSelf()
      channelStore.channels.forEach(channel => {
        const userDataRequest = packers[MessageTypes.UserDataRequest]({ username: self.name, colour: self.color, clientId: CLIENT_ID }, channel.key.raw)
        ws.send(userDataRequest)
      })
    }
    
    app.provide('requestUserData', requestUserData)

    // Keepalive timeout.
    var timeout = setInterval(() => {
      ws.send(packers[MessageTypes.Keepalive]())
    }, 30000)

    setInterval(() => {
      requestUserData()
    }, 30000)
 
    ws.addEventListener('open', () => {
      // Set global connection state for UI.
      mercuryStore.setConnectionState(true)

      // Subscribe to all messages.
      SUBSCRIBED_MESSAGE_TYPES.forEach((messageType) => {
        ws.send(packers[MessageTypes.Subscribe]({ messageType: messageType }))
      })

      // Make a user data request for each channel.
      requestUserData()
    })

    ws.addEventListener('close', () => {
      mercuryStore.setConnectionState(false)
    })

    ws.addEventListener('message', (event) => {
      const incomingPacket = unpackIncomingPacket(new Uint8Array(event.data))

      // Update user store.
      userStore.updateUser({id: incomingPacket.senderId})

      // Parse packet using message type.
      switch (incomingPacket.messageType) {
        case MessageTypes.Basic:
          handleBasicMessage(incomingPacket)
          break;
        case MessageTypes.UserDataRequest:
          handleUserDataRequest(incomingPacket)
          break;
        case MessageTypes.UserDataResponse:
          handleUserDataResponse(incomingPacket)
          break;

        default:
          console.error(`Received unknown message type ${incomingPacket.messageType}`);
      }

      // Restart keepalive.
      clearInterval(timeout)
      timeout = setInterval(() => {
        console.log("Keepalive message sent!")
        ws.send(packers[MessageTypes.Keepalive]())
      }, 30000)
    })
  }
}
